package lsgui.gui;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import javax.swing.JOptionPane;
import javax.swing.table.DefaultTableModel;
import lsgui.Main;

public class AnchorPositionDialog extends javax.swing.JDialog {
    
    private AnchorTableModel model;
    
    /** Creates new form AnchorPositionDialog */
    public AnchorPositionDialog(java.awt.Frame parent, boolean modal,
            Canvas canvas) {
        super(parent, modal);
        initComponents();
        model = new AnchorTableModel(canvas);
        anchorTable.setModel(model);
        setLocationRelativeTo(parent);
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        pAnchorTable = new javax.swing.JPanel();
        spAnchorTable = new javax.swing.JScrollPane();
        anchorTable = new javax.swing.JTable();
        jPanel1 = new javax.swing.JPanel();
        bLoad = new javax.swing.JButton();
        bSave = new javax.swing.JButton();
        bClose = new javax.swing.JButton();

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        setTitle("Edit anchor locations");
        setResizable(false);

        pAnchorTable.setBorder(javax.swing.BorderFactory.createEmptyBorder(5, 5, 0, 5));
        pAnchorTable.setLayout(new java.awt.BorderLayout());

        spAnchorTable.setViewportView(anchorTable);

        pAnchorTable.add(spAnchorTable, java.awt.BorderLayout.CENTER);

        bLoad.setText("Load...");
        bLoad.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bLoadActionPerformed(evt);
            }
        });
        jPanel1.add(bLoad);

        bSave.setText("Save...");
        bSave.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bSaveActionPerformed(evt);
            }
        });
        jPanel1.add(bSave);

        bClose.setText("Close");
        bClose.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bCloseActionPerformed(evt);
            }
        });
        jPanel1.add(bClose);

        pAnchorTable.add(jPanel1, java.awt.BorderLayout.SOUTH);

        getContentPane().add(pAnchorTable, java.awt.BorderLayout.CENTER);

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void bCloseActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bCloseActionPerformed
        dispose();
    }//GEN-LAST:event_bCloseActionPerformed

    private void bSaveActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bSaveActionPerformed
        File file = Main.showSaveFileDialog(getParent(), "Save anchors",
                null, Main.PROPERTY_ANCHOR_SAVE_PATH);
        if (file != null) {
            String filename = file.getName();
            if (!filename.endsWith(".ser")) {
                filename += ".ser";
            }
            File sFile = new File(file.getParentFile(), filename);
            if (sFile.exists()) {
                int res = JOptionPane.showConfirmDialog(getParent(), "A file with"
                        + " this name already exists. Do you\nwant to overwrite it?",
                        "Overwrite file?", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                if (res != JOptionPane.YES_OPTION) {
                    return;
                }
            }
            boolean success = model.saveToFile(sFile);
            if (!success) {
                JOptionPane.showMessageDialog(getParent(), "Error while saving"
                        + " anchor locations to file!", "Error", JOptionPane.ERROR_MESSAGE);
            }
        }
    }//GEN-LAST:event_bSaveActionPerformed

    private void bLoadActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bLoadActionPerformed
        File file = Main.showOpenFileDialog(getParent(), "Load anchors",
                new String[] {"Serialized Anchor File (*.ser)"},
                new String[][] {{"ser", "SER"}},
                Main.PROPERTY_ANCHOR_OPEN_PATH);
        if (file != null) {
            boolean success = model.loadFromFile(file);
            if (!success) {
                JOptionPane.showMessageDialog(getParent(), "Error while loading"
                        + " anchor locations from file!", "Error", JOptionPane.ERROR_MESSAGE);
            }
        }
    }//GEN-LAST:event_bLoadActionPerformed

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JTable anchorTable;
    private javax.swing.JButton bClose;
    private javax.swing.JButton bLoad;
    private javax.swing.JButton bSave;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel pAnchorTable;
    private javax.swing.JScrollPane spAnchorTable;
    // End of variables declaration//GEN-END:variables

    
    private static class AnchorTableModel extends DefaultTableModel {
        
        private Canvas canvas;
        private ArrayList<java.awt.Point> anchors;
        
        private static final String[] headers = new String[] {"Anchor-ID", "x", "y"};
        
        public AnchorTableModel(Canvas canvas) {
            this.canvas = canvas;
            this.anchors = canvas.getAnchors();
        }

        @Override
        public int getRowCount() {
            return (anchors != null) ? anchors.size() : 0;
        }

        @Override
        public int getColumnCount() {
            return headers.length;
        }

        @Override
        public String getColumnName(int i) {
            return headers[i];
        }

        @Override
        public Class<?> getColumnClass(int i) {
            return Integer.class;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column > 0;
        }

        @Override
        public Object getValueAt(int row, int column) {
            switch (column) {
                case 0:
                    return row + 1;
                case 1:
                    return anchors.get(row).x;
                case 2:
                    return anchors.get(row).y;
                default:
                    return -1;
            }
        }

        @Override
        public void setValueAt(Object o, int row, int column) {
            int value = (Integer)o;
            java.awt.Point p = anchors.get(row);
            if (column == 1) {
                if (value < 0 || value >= canvas.getCanvasWidth()) {
                    return;
                }
                p.x = value;
                canvas.repaintRequest();
            } else {
                if (value < 0 || value >= canvas.getCanvasHeight()) {
                    return;
                }
                p.y = value;
                canvas.repaintRequest();
            }
        }
        
        private boolean saveToFile(File file) {
            try {
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                new ObjectOutputStream(baos).writeObject(anchors);
                return Main.saveByteArray(baos.toByteArray(), file);
            } catch (IOException e) {
                return false;
            }
        }

        private boolean loadFromFile(File file) {
            try {
                FileInputStream fis = new FileInputStream(file);
                ObjectInputStream o = new ObjectInputStream(fis);
                anchors = (ArrayList<java.awt.Point>) o.readObject();
                if (!canvas.checkAnchors(anchors)) {
                    o.close();
                    return false;
                }
                canvas.setAnchors(anchors);
                fireTableDataChanged();
                o.close();
                return true;
            } catch (IOException e) {
                System.err.println(e);
            } catch (ClassNotFoundException e) {
                System.err.println(e);
            }
            return false;
        }
        
    }
    
}
