package lsgui.gui;

import com.sun.jna.Memory;
import com.sun.jna.Pointer;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import javax.imageio.ImageIO;
import javax.swing.ButtonGroup;
import javax.swing.JOptionPane;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTabbedPane;
import latmath.algorithm.BasicLaterationAlgorithm;
import latmath.algorithm.LaterationAlgorithmCollection;
import latmath.algorithm.LowerBoundEstimator;
import latmath.errormodel.ErrorModel;
import latmath.errormodel.ErrorModelLos;
import latmath.errormodel.ErrorModelLosNlos;
import latmath.errormodel.ErrorModelLosNlosUniform;
import latmath.errormodel.ErrorModelNanopan5375;
import latmath.errormodel.ErrorModelUniform;
import latmath.util.Point2d;
import lsgui.Main;
import lsgui.jna.data.LS2Result;
import lsgui.jna.data.ShooterTGAImage;
import lsgui.util.LaterationImage;
import lsgui.verify.OutputVariant;
import lsgui.verify.Verifier;

public class MainFrame extends javax.swing.JFrame {

    private volatile boolean cancelVerify;

    private DecimalFormat format = new DecimalFormat("0.000 sec");

    private static final int TAB_LIMIT = 16;
    private static final String NEW_TAB = "New Tab";

    /** List of available error models for verify method */
    private int errorModelIdx = 0;
    private ErrorModel[] errorModels;
    private ButtonGroup emBgrp;
    private JRadioButtonMenuItem currentErrorModelControl;
    
    public static boolean drawAnchors = true;

    /** Creates new form MainFrame */
    public MainFrame() {
        initData();
        initComponents();
        spWidth.setModel(new javax.swing.SpinnerNumberModel(LS2Result.WIDTH_DEFAULT, 100, 1000, 1));
        spHeight.setModel(new javax.swing.SpinnerNumberModel(LS2Result.HEIGHT_DEFAULT, 100, 1000, 1));
        addTab(false);
        lCPUInfo.setText("CPUs: " + Runtime.getRuntime().availableProcessors());
        cbAlgorithm.setModel(new javax.swing.DefaultComboBoxModel(Main.algorithms));
        cbErrorModel.setModel(new javax.swing.DefaultComboBoxModel(Main.errorModels));
        java.awt.Dimension screenSize = java.awt.Toolkit.getDefaultToolkit().getScreenSize();
        int cw = (int)(screenSize.width/1.5);
        int ch = (int)(screenSize.height/1.5);
        setBounds((screenSize.width-cw)/2, (screenSize.height-ch)/2, cw, ch);

        // Add error model selection to GUI
        emBgrp = new ButtonGroup();
        for (int i = 0; i < errorModels.length; i++) {
            JRadioButtonMenuItem menuItem = new JRadioButtonMenuItem(errorModels[i].toString());
            menuItem.setActionCommand("" + i);
            menuItem.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    errorModelIdx = Integer.parseInt(e.getActionCommand());
                    miConfigureErrorModel.setEnabled(errorModels[errorModelIdx].isConfigurable());
                    currentErrorModelControl = (JRadioButtonMenuItem) e.getSource();
                }
            });
            menuItem.setSelected(i == 0);
            mErrorModel.add(menuItem);
            emBgrp.add(menuItem);
        }

        if (Main.INSTANCE == null) {
            showNoSSELibraryError();
        }
    }

    public JTabbedPane getTabbedPane() {
        return tabbedPane;
    }
    
    private void initData() {
        errorModels = new ErrorModel[5];
        errorModels[0] = new ErrorModelUniform(100, 0, 100);
        errorModels[0].setNegativeOffsetsEnabled(true);
        errorModels[1] = new ErrorModelNanopan5375(Double.MAX_VALUE);
        errorModels[1].setScaleFactor(ShooterTGAImage.VERY_GOOD_VALUES_COLOR / 2.85);
        errorModels[1].setNegativeOffsetsEnabled(true);
        errorModels[2] = new ErrorModelLosNlos(1000.0, 5.0, 2.5, 2.38419e-07, 3);
        errorModels[2].setScaleFactor(ShooterTGAImage.VERY_GOOD_VALUES_COLOR / 6.8);
        errorModels[2].setNegativeOffsetsEnabled(true);
        errorModels[3] = new ErrorModelLosNlosUniform(1000.0, 5.0, 2.5, 0, 0);
        errorModels[3].setScaleFactor(1.5 * ShooterTGAImage.VERY_GOOD_VALUES_COLOR / 6.8);
        errorModels[3].setNegativeOffsetsEnabled(true);
        errorModels[4] = new ErrorModelLos(1000.0, 0.0, 30.0);
        errorModels[4].setNegativeOffsetsEnabled(true);
    }

    private int addTab(boolean clone) {
        if (tabbedPane.getTabCount() >= TAB_LIMIT) {
            JOptionPane.showMessageDialog(this, "Maximum number of tabs reached!",
                    "Information", JOptionPane.INFORMATION_MESSAGE);
            return -1;
        }

        Point[] copy = null;
        int canvasWidth = (Integer) spWidth.getValue();
        int canvasHeight = (Integer) spHeight.getValue();

        if (clone) {
            Canvas c = getCurrentCanvas();
            canvasWidth = c.getCanvasWidth();
            canvasHeight = c.getCanvasHeight();
            ArrayList<Point> list = c.getAnchors();
            if (list.size() > 0) {
                copy = new Point[list.size()];
                for (int i = 0; i < list.size(); i++) {
                    copy[i] = new Point(list.get(i).x, list.get(i).y);
                }
            }
        }

        ScrollableCanvas sCanvas = new ScrollableCanvas(canvasWidth,
                canvasHeight);
        TabContent tabContent = new TabContent(this, sCanvas);
        if (clone && copy != null) {
            sCanvas.getCanvas().setAnchorsWOR(copy);
        }
        tabbedPane.addTab(NEW_TAB, tabContent);
        int count = tabbedPane.getTabCount();
        tabbedPane.setTabComponentAt(count-1, new ButtonTabComponent(tabbedPane));
        tabbedPane.setSelectedIndex(count-1);
        return count-1;
    }
    
    private void saveTab(boolean all) {
        Canvas c = getCurrentCanvas();
        LaterationImage image = c.getLatImage();
        if (image == null || !image.hasImages()) {
            JOptionPane.showMessageDialog(MainFrame.this,
                    "Nothing to save, please start a simulation!", "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        
        File file = Main.showSaveFileDialog(this, "Set image filename",
                getCurrentTabTitle().replaceAll(":", "").replaceAll(" ", "_"),
                Main.PROPERTY_TAB_SAVE_PATH);
        if (file != null) {
            String filename = file.getName();
            if (filename.endsWith(".png")) {
                int idx = filename.lastIndexOf(".png");
                filename = filename.substring(0, idx);
            }
            
            int size = all ? image.getImageCount() : 1;
            File[] files = new File[size];
            String[] fileEndings = new String[size];
            BufferedImage[] images = new BufferedImage[size];
            if (all) {
                int savedImageIndex = image.getImageIndex();
                for (int i = 0; i < size; i++) {
                    image.setActiveImage(i);
                    images[i] = image.getRenderedImage().image;
                    fileEndings[i] = image.getLaterationImageData().getFileTag();
                }
                image.setActiveImage(savedImageIndex);
            } else {
                images[0] = image.getRenderedImage().image;
                fileEndings[0] = image.getLaterationImageData().getFileTag();
            }
            
            try {
                boolean exists = false;
                for (int i = 0; i < fileEndings.length; i++) {
                    files[i] = new File(file.getParentFile(), filename + "_" + fileEndings[i] + ".png");
                    exists |= files[i].exists();
                }
                
                if (exists) {
                    int res = JOptionPane.showConfirmDialog(this, "A file with"
                            + " this name already exists. Do you\nwant to overwrite it?",
                            "Overwrite file?", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                    if (res != JOptionPane.YES_OPTION) {
                        return;
                    }
                }
                
                for (int i = 0; i < fileEndings.length; i++) {
                    c.drawAnchors(images[i].createGraphics());
                    ImageIO.write(images[i], "png", files[i]);
                }
            } catch (Exception e) {
                JOptionPane.showMessageDialog(MainFrame.this,
                        e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }
        }
    }

    private String getCurrentTabTitle() {
        int idx = tabbedPane.getSelectedIndex();
        return tabbedPane.getTitleAt(idx);
    }
    
    private void setCurrentTabTitle(String s) {
        int idx = tabbedPane.getSelectedIndex();
        tabbedPane.setTitleAt(idx, s);
        tabbedPane.getTabComponentAt(idx).invalidate();
    }

    private TabContent getCurrentTabContent() {
        return ((TabContent)tabbedPane.getSelectedComponent());
    }
    
    private Canvas getCurrentCanvas() {
        TabContent tabContent = getCurrentTabContent();
        return tabContent.getScrollableCanvas().getCanvas();
    }

    private int getRUNS() {
        return (Integer) spRuns.getValue();
    }
    
    private void lock() {
        Main.operationPending = true;
    }
    
    private void unlock() {
        Main.operationPending = false;
    }
    
    public void start() {
        if (Main.INSTANCE == null) {
            showNoSSELibraryError();
            return;
        }
        
        if (bStart.getText().equalsIgnoreCase("cancel")) {
            cancelRun();
            return;
        }
        
        // if any operation pending, don't be able to start a new one
        if (Main.operationPending) {
            return;
        }
        
        final Canvas canvas = getCurrentCanvas();
        final TabContent tabContent = getCurrentTabContent();
        ArrayList<Point> anchors = canvas.getAnchors();
        if (anchors.size() < 3) {
            showInsufficientAnchorsError();
            return;
        }
        final int runs = getRUNS();
        final LS2Result result = new LS2Result(canvas.getCanvasWidth(),
                canvas.getCanvasHeight());
        OutputVariantDialog ovDialog = new OutputVariantDialog(this, true,
                Main.outputVariants, result);
        ovDialog.setVisible(true);
        
        boolean allNull = true;
        for (int i = 0; i < result.data.length; i++) {
            if (result.data[i] != null) {
                allNull = false;
                break;
            }
        }
        
        if (allNull) {
            return;
        }

        (new Thread() {
            @Override
            public void run() {
                lock();
                int i = tabbedPane.getSelectedIndex();
                setButtonsEnabled(canvas, i, false);
                bStart.setText("Cancel");
                bStart.setEnabled(true);
                setCurrentTabTitle((String)cbAlgorithm.getSelectedItem());
                long time = runThreaded(runs, result);
                if (time != -1) {
                    setRuntime(tabContent, time);
                }
                unlock();
                setButtonsEnabled(canvas, i, true);
                bStart.setText("Start");
            }
        }).start();
    }

    public void verify() {

        if (bVerify.getText().equalsIgnoreCase("cancel")) {
            cancelVerify();
            return;
        }
        
        // if any operation pending, don't be able to start a new one
        if (Main.operationPending) {
            return;
        }

        final Canvas canvas = getCurrentCanvas();
        final TabContent tabContent = getCurrentTabContent();
        ArrayList<Point> anchors = canvas.getAnchors();
        if (anchors.size() < 3) {
            showInsufficientAnchorsError();
            return;
        }
        // find algorithm in latmath package that matches selected algorithm
        // by name, show dialog and pre-select if found
        String name = (String) cbAlgorithm.getSelectedItem();
        List<BasicLaterationAlgorithm> algs = LaterationAlgorithmCollection.getAlgorithms();
        BasicLaterationAlgorithm[] algArray = algs.toArray(new BasicLaterationAlgorithm[0]);
        VerifyDialog dialog = new VerifyDialog(this, true, name, algArray);
        dialog.setVisible(true);
        
        BasicLaterationAlgorithm selected = dialog.getSelectedAlgorithm();
        if (selected == null) {
            return;
        }
        
        // determine output variants
        final LS2Result result = new LS2Result(canvas.getCanvasWidth(),
                canvas.getCanvasHeight());
        OutputVariantDialog ovDialog = new OutputVariantDialog(this, true,
                OutputVariant.toStringArray(), result);
        ovDialog.setVisible(true);
        
        boolean allNull = true;
        for (int i = 0; i < result.data.length; i++) {
            if (result.data[i] != null) {
                allNull = false;
                break;
            }
        }
        
        if (allNull) {
            return;
        }
        
        final int runs = getRUNS();
        final BasicLaterationAlgorithm taken = selected;

        // find error model to use
        final ErrorModel emTaken = errorModels[errorModelIdx];

        (new Thread() {
            @Override
            public void run() {
                lock();
                cancelVerify = false;
                int i = tabbedPane.getSelectedIndex();
                setButtonsEnabled(canvas, i, false);
                bVerify.setText("Cancel");
                bVerify.setEnabled(true);
                setCurrentTabTitle("Verify: " + taken.getName());
                long time = runVerifyThreaded(runs, taken, emTaken, result);
                if (time != -1) {
                    setRuntime(tabContent, time);
                }
                unlock();
                setButtonsEnabled(canvas, i, true);
                bVerify.setText("Verify");
            }
        }).start();
    }
    
    public void compare(final LaterationImage img1, final int idx1, final LaterationImage img2, final int idx2) {
        
        // if any operation pending, don't be able to start a new one
        if (Main.operationPending) {
            return;
        }
        
        if (img1.getImageWidth() != img2.getImageWidth() ||
                img1.getImageHeight() != img2.getImageHeight() ||
                img1.getEngine() != img2.getEngine()) {
            return;
        }
        int newTabIndex = addTab(true);
        if (newTabIndex != -1) {
            final Canvas canvas = getCurrentCanvas();
            final TabContent tabContent = getCurrentTabContent();
            setCurrentTabTitle("Compare: " + img1.getAlgorithmName()
                    + " vs. " + img2.getAlgorithmName());
            (new Thread() {
                @Override
                public void run() {
                    lock();
                    setButtonsEnabled(canvas, 0, false);
                    long time = runCompareThreaded(img1, idx1, img2, idx2);
                    if (time != -1) {
                        setRuntime(tabContent, time);
                    }
                    unlock();
                    setButtonsEnabled(canvas, 0, true);
                }
            }).start();
        }
    }

    private void cancelRun() {
        int res, count = 0;
        do {
            try { Thread.sleep(10); } catch (InterruptedException ex) {}
            res = Main.INSTANCE.cancel_running();
            count++;
        } while (res != 1 && count < 25); // possibly not more than 250 ms for starting worker threads
    }

    private long runThreaded(int runs, LS2Result result) {
        Canvas canvas = getCurrentCanvas();
        ArrayList<Point> anchors = canvas.getAnchors();
        if (anchors.size() < 3) {
            return -1;
        }
        int size = anchors.size();
        float[] x = new float[size];
        float[] y = new float[size];
        for (int i = 0; i < size; i++) {
            x[i] = anchors.get(i).x;
            y[i] = anchors.get(i).y;
        }
        
        // conversion: prepare result data
        Pointer[] paList = new Pointer[result.data.length];
        for (int i = 0; i < paList.length; i++) {
            if (result.data[i] != null) {
                paList[i] = new Memory(result.data[i].length * (Float.SIZE/8));
            }
        }
        
        long elapsed = System.currentTimeMillis();
        size = Main.INSTANCE.compute_locbased(
                cbAlgorithm.getSelectedIndex(),
                cbErrorModel.getSelectedIndex(),
                Runtime.getRuntime().availableProcessors(),
                runs,
                x,
                y,
                size,
                paList,
                result.width,
                result.height
                );
        elapsed = System.currentTimeMillis() - elapsed;
        
        // conversion: get result data back from pointer array
        for (int i = 0; i < paList.length; i++) {
            if (result.data[i] != null) {
                paList[i].read(0, result.data[i], 0, result.data[i].length);
            }
        }
        
        if (size >= 0) {
            LaterationImage img = new LaterationImage(result,
                    LaterationImage.ENGINE_LIBRARY,
                    ShooterTGAImage.IMAGE_MODE_ONE);
            img.setAlgorithmName((String)cbAlgorithm.getSelectedItem());
            img.setAlgorithmType(LaterationImage.TYPE_LOCATION_ALGORITHM);
            canvas.setLatImage(img);
        }
        return elapsed;
    }

    private void cancelVerify() {
        cancelVerify = true;
    }

    private long runVerifyThreaded(int runs, BasicLaterationAlgorithm alg,
            ErrorModel emTaken, LS2Result result) {
        Canvas canvas = getCurrentCanvas();
        ArrayList<Point> anchors = canvas.getAnchors();
        if (anchors.size() < 3) {
            return -1;
        }
        int size = anchors.size();

        Point2d[] points = new Point2d[size];
        for (int i = 0; i < size; i++) {
            points[i] = new Point2d(anchors.get(i).x, anchors.get(i).y);
        }

        int processors = Runtime.getRuntime().availableProcessors();
        int slice = result.height / processors;
        int sliceLastAdditional = result.height % processors;
        Thread[] threads = new Thread[processors];
        Verifier[] verifiers = new Verifier[processors];
        for (int i = 0; i < processors; i++) {
            if (i != processors - 1) {
                verifiers[i] = new Verifier(runs, i*slice, slice, result,
                        (ErrorModel)emTaken.clone(), alg, points,
                        Thread.currentThread());
            } else {
                verifiers[i] = new Verifier(runs, i*slice,
                        slice+sliceLastAdditional, result,
                        (ErrorModel)emTaken.clone(), alg, points,
                        Thread.currentThread());
            }
            threads[i] = new Thread(verifiers[i]);
        }

        long elapsed = System.currentTimeMillis();
        for (int i = 0; i < processors; i++) {
            threads[i].start();
        }

        while (!cancelVerify) {
            int progress = 0;
            for (int i = 0; i < processors; i++) {
                progress += verifiers[i].getCurrentStep();
            }
            lpercent.setText("Verify in progress: "
                    + ((progress * 100) / result.height) + "% complete"
                    + ", time elapsed: " + format.format(
                    (System.currentTimeMillis() - elapsed) / 1000.0));
            if (progress == result.height) {
                break;
            }
            try { Thread.sleep(500); } catch (InterruptedException ex) {}
        }

        elapsed = System.currentTimeMillis() - elapsed;
        lpercent.setText(null);

        if (cancelVerify) {
            for (Verifier v : verifiers) {
                v.stop();
            }
            for (Thread t : threads) {
                try { t.join(); } catch (InterruptedException ex) {}
            }
        }
        
        LaterationImage img = new LaterationImage(result,
                LaterationImage.ENGINE_VERIFY, ShooterTGAImage.IMAGE_MODE_ONE);
        img.setAlgorithmName(alg.getName());
        if (alg instanceof LowerBoundEstimator) {
            img.setAlgorithmType(LaterationImage.TYPE_LOWER_BOUND_ESTIMATOR);
        } else {
            img.setAlgorithmType(LaterationImage.TYPE_LOCATION_ALGORITHM);
        }
        canvas.setLatImage(img);
        return elapsed;
    }

    private long runCompareThreaded(LaterationImage img1, int idx1, LaterationImage img2, int idx2) {
        Canvas canvas = getCurrentCanvas();
        long elapsed = System.currentTimeMillis();
        // create result
        LS2Result result = new LS2Result(img1.getImageWidth(),
                img1.getImageHeight());
        // and store values at same output variant index as image one
        result.initData(idx1+1);
        result.initDataAt(idx1);
        // it is fast enough to run compare single threaded...
        float[] img1Data = img1.getLaterationImageData(idx1).getImageData();
        float[] img2Data = img2.getLaterationImageData(idx2).getImageData();
        for (int i = 0; i < img1Data.length; i++) {
            if (LS2Result.isError(img1Data[i]) || LS2Result.isError(img2Data[i])) {
                result.data[idx1][i] = LS2Result.ERROR;
            } else {
                result.data[idx1][i] = img1Data[i] - img2Data[i];
            }
        }
        // create final lateration image, choose image mode...
        int imageMode = ShooterTGAImage.IMAGE_MODE_TWO;
        if (img1.getAlgorithmType() != img2.getAlgorithmType()) {
            imageMode = ShooterTGAImage.IMAGE_MODE_THREE;
        }
        LaterationImage img = new LaterationImage(result,
                img1.getEngine(), imageMode);
        elapsed = System.currentTimeMillis() - elapsed;
        canvas.setLatImage(img);
        return elapsed;
    }
    
    public void reset() {
        getCurrentCanvas().reset();
        setCurrentTabTitle(NEW_TAB);
        lruntime.setText(null);
    }

    public void setRuntime(TabContent tabContent, long millis) {
        tabContent.setRuntime(format.format(millis/1000.0));
    }

    public void setButtonsEnabled(Canvas canvas, int tab, boolean enabled) {
        canvas.setEventsOn(enabled);
        cbAlgorithm.setEnabled(enabled);
        bStart.setEnabled(enabled);
        bVerify.setEnabled(enabled);
        bReset.setEnabled(enabled);
        bAddTab.setEnabled(enabled);
        spRuns.setEnabled(enabled);
        spWidth.setEnabled(enabled);
        spHeight.setEnabled(enabled);
        cbErrorModel.setEnabled(enabled);
        for (int i = 0; i < tabbedPane.getTabCount(); i++) {
            TabContent tabC = (TabContent) tabbedPane.getComponentAt(i);
            tabC.setControlsEnabled(enabled);
        }
    }
    
    private void showNoSSELibraryError() {
        JOptionPane.showMessageDialog(this, "<html>The native library wasn't"
                + " found in the default locations. Please copy the library to<br>"
                + " the application directory if you want to use fast SSE/AVX support"
                + " and then restart.</html>", "No native library",
                JOptionPane.ERROR_MESSAGE);
    }
    
    private void showInsufficientAnchorsError() {
        JOptionPane.showMessageDialog(this, "<html>Please add at least three"
                + " anchors to the playing field (left click).</html>",
                "Insufficient anchors", JOptionPane.ERROR_MESSAGE);
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        tabPopupMenu = new javax.swing.JPopupMenu();
        miTabBlank = new javax.swing.JMenuItem();
        miTabClone = new javax.swing.JMenuItem();
        mTabSave = new javax.swing.JMenu();
        miTabSave = new javax.swing.JMenuItem();
        miTabSaveAll = new javax.swing.JMenuItem();
        miTabCloseAll = new javax.swing.JMenuItem();
        jSeparator1 = new javax.swing.JPopupMenu.Separator();
        mErrorModel = new javax.swing.JMenu();
        miConfigureErrorModel = new javax.swing.JMenuItem();
        miTestErrorModel = new javax.swing.JMenuItem();
        jSeparator2 = new javax.swing.JPopupMenu.Separator();
        miDrawAnchors = new javax.swing.JCheckBoxMenuItem();
        miEditAnchors = new javax.swing.JMenuItem();
        toolBarContainer = new javax.swing.JPanel();
        toolbar = new javax.swing.JPanel();
        lCPUInfo = new javax.swing.JLabel();
        jLabel1 = new javax.swing.JLabel();
        spRuns = new javax.swing.JSpinner();
        lSelectInfo = new javax.swing.JLabel();
        cbAlgorithm = new javax.swing.JComboBox();
        jLabel2 = new javax.swing.JLabel();
        cbErrorModel = new javax.swing.JComboBox();
        toolbar2 = new javax.swing.JPanel();
        bAddTab = new javax.swing.JButton();
        jLabel3 = new javax.swing.JLabel();
        spWidth = new javax.swing.JSpinner();
        jLabel4 = new javax.swing.JLabel();
        spHeight = new javax.swing.JSpinner();
        bStart = new javax.swing.JButton();
        bVerify = new javax.swing.JButton();
        bReset = new javax.swing.JButton();
        lruntime = new javax.swing.JLabel();
        lpercent = new javax.swing.JLabel();
        tabbedPane = new javax.swing.JTabbedPane();

        miTabBlank.setText("New Tab");
        miTabBlank.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miTabBlankActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miTabBlank);

        miTabClone.setText("Clone Tab");
        miTabClone.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miTabCloneActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miTabClone);

        mTabSave.setText("Save Tab");

        miTabSave.setText("Current Image...");
        miTabSave.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miTabSaveActionPerformed(evt);
            }
        });
        mTabSave.add(miTabSave);

        miTabSaveAll.setText("All Images...");
        miTabSaveAll.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miTabSaveAllActionPerformed(evt);
            }
        });
        mTabSave.add(miTabSaveAll);

        tabPopupMenu.add(mTabSave);

        miTabCloseAll.setText("Close Tabs");
        miTabCloseAll.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miTabCloseAllActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miTabCloseAll);
        tabPopupMenu.add(jSeparator1);

        mErrorModel.setText("Error model");
        tabPopupMenu.add(mErrorModel);

        miConfigureErrorModel.setText("Configure error model...");
        miConfigureErrorModel.setEnabled(false);
        miConfigureErrorModel.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miConfigureErrorModelActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miConfigureErrorModel);

        miTestErrorModel.setText("Test error model...");
        miTestErrorModel.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miTestErrorModelActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miTestErrorModel);
        tabPopupMenu.add(jSeparator2);

        miDrawAnchors.setSelected(true);
        miDrawAnchors.setText("Draw anchors");
        miDrawAnchors.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miDrawAnchorsActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miDrawAnchors);

        miEditAnchors.setText("Edit anchors...");
        miEditAnchors.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                miEditAnchorsActionPerformed(evt);
            }
        });
        tabPopupMenu.add(miEditAnchors);

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("Lateration Shooter GUI");
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                formWindowClosing(evt);
            }
        });

        toolBarContainer.setLayout(new javax.swing.BoxLayout(toolBarContainer, javax.swing.BoxLayout.PAGE_AXIS));

        toolbar.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 10, 5));

        lCPUInfo.setText("jLabel1");
        toolbar.add(lCPUInfo);

        jLabel1.setText("Runs:");
        toolbar.add(jLabel1);

        spRuns.setModel(new javax.swing.SpinnerNumberModel(40, 1, 9999, 1));
        toolbar.add(spRuns);

        lSelectInfo.setText("Algorithm:");
        toolbar.add(lSelectInfo);
        toolbar.add(cbAlgorithm);

        jLabel2.setText("Error model:");
        toolbar.add(jLabel2);
        toolbar.add(cbErrorModel);

        toolBarContainer.add(toolbar);

        toolbar2.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 10, 5));

        bAddTab.setText("+");
        bAddTab.setToolTipText("Tab actions");
        bAddTab.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bAddTabActionPerformed(evt);
            }
        });
        toolbar2.add(bAddTab);

        jLabel3.setText("Size:");
        toolbar2.add(jLabel3);
        toolbar2.add(spWidth);

        jLabel4.setText("x");
        toolbar2.add(jLabel4);
        toolbar2.add(spHeight);

        bStart.setText("Start");
        bStart.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bStartActionPerformed(evt);
            }
        });
        toolbar2.add(bStart);

        bVerify.setText("Verify");
        bVerify.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bVerifyActionPerformed(evt);
            }
        });
        toolbar2.add(bVerify);

        bReset.setText("Reset");
        bReset.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bResetActionPerformed(evt);
            }
        });
        toolbar2.add(bReset);
        toolbar2.add(lruntime);
        toolbar2.add(lpercent);

        toolBarContainer.add(toolbar2);

        getContentPane().add(toolBarContainer, java.awt.BorderLayout.NORTH);

        tabbedPane.setTabLayoutPolicy(javax.swing.JTabbedPane.SCROLL_TAB_LAYOUT);
        getContentPane().add(tabbedPane, java.awt.BorderLayout.CENTER);

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void bStartActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bStartActionPerformed
        start();
    }//GEN-LAST:event_bStartActionPerformed

    private void bResetActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bResetActionPerformed
        reset();
    }//GEN-LAST:event_bResetActionPerformed

    private void bVerifyActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bVerifyActionPerformed
        verify();
    }//GEN-LAST:event_bVerifyActionPerformed

    private void bAddTabActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bAddTabActionPerformed
        tabPopupMenu.show(bAddTab, 0, bAddTab.getHeight());
    }//GEN-LAST:event_bAddTabActionPerformed

    private void miTabBlankActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miTabBlankActionPerformed
        addTab(false);
    }//GEN-LAST:event_miTabBlankActionPerformed

    private void miTabCloneActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miTabCloneActionPerformed
        addTab(true);
    }//GEN-LAST:event_miTabCloneActionPerformed

    private void miTabSaveActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miTabSaveActionPerformed
        saveTab(false);
    }//GEN-LAST:event_miTabSaveActionPerformed

    private void miEditAnchorsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miEditAnchorsActionPerformed
        final AnchorPositionDialog dialog = new AnchorPositionDialog(this, true,
                getCurrentCanvas());
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                dialog.setVisible(true);
            }
        });
    }//GEN-LAST:event_miEditAnchorsActionPerformed

    private void miTestErrorModelActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miTestErrorModelActionPerformed
        final ErrorModel model = errorModels[errorModelIdx];
        JOptionPane.showMessageDialog(this, "<html>1.000.000 samples will be drawn"
                + " using \""+model+"\"<br>error model. Please set the filename in the next step.</html>",
                "Information", JOptionPane.INFORMATION_MESSAGE);
        final File file = Main.showSaveFileDialog(this, "Save file", "samples.log", Main.PROPERTY_EM_DATA_SAVE_PATH);
        if (file != null) {
            PrintWriter out = null;
            //final Canvas canvas = this.getCurrentCanvas();
            //setButtonsEnabled(canvas, tabbedPane.getSelectedIndex(), false);
            try {
                FileWriter outFile = new FileWriter(file);
                out = new PrintWriter(outFile);
                for (int i = 0; i < 1000000; i++) {
                    double sample = model.getOffset(0, 0);
                    out.println(sample);
                    if (i % 100 == 0) {
                        out.flush();
                    }
                }
                out.flush();
            } catch (IOException e) {
                System.err.println(e.getMessage());
            } finally {
                out.close();
            }
            //setButtonsEnabled(canvas, tabbedPane.getSelectedIndex(), true);
        }
    }//GEN-LAST:event_miTestErrorModelActionPerformed

    private void miConfigureErrorModelActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miConfigureErrorModelActionPerformed
        ErrorModel model = errorModels[this.errorModelIdx];
        if (model.isConfigurable()) {
            model.configure(this);
            if (currentErrorModelControl != null) {
                currentErrorModelControl.setText(errorModels[this.errorModelIdx].toString());
            }
        }
    }//GEN-LAST:event_miConfigureErrorModelActionPerformed

    private void miDrawAnchorsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miDrawAnchorsActionPerformed
        drawAnchors = miDrawAnchors.isSelected();
        for (int i = 0; i < tabbedPane.getTabCount(); i++) {
            TabContent tabContent = ((TabContent)tabbedPane.getComponentAt(i));
            Canvas canvas = tabContent.getScrollableCanvas().getCanvas();
            canvas.repaintRequest();
        }
    }//GEN-LAST:event_miDrawAnchorsActionPerformed

    private void formWindowClosing(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosing
        Main.exit();
    }//GEN-LAST:event_formWindowClosing

    private void miTabSaveAllActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miTabSaveAllActionPerformed
        saveTab(true);
    }//GEN-LAST:event_miTabSaveAllActionPerformed

    private void miTabCloseAllActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_miTabCloseAllActionPerformed
        for (int i = tabbedPane.getTabCount() - 1; i > 0; i--) {
            tabbedPane.remove(i);
        }
    }//GEN-LAST:event_miTabCloseAllActionPerformed

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton bAddTab;
    private javax.swing.JButton bReset;
    private javax.swing.JButton bStart;
    private javax.swing.JButton bVerify;
    private javax.swing.JComboBox cbAlgorithm;
    private javax.swing.JComboBox cbErrorModel;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JLabel jLabel4;
    private javax.swing.JPopupMenu.Separator jSeparator1;
    private javax.swing.JPopupMenu.Separator jSeparator2;
    private javax.swing.JLabel lCPUInfo;
    private javax.swing.JLabel lSelectInfo;
    private javax.swing.JLabel lpercent;
    private javax.swing.JLabel lruntime;
    private javax.swing.JMenu mErrorModel;
    private javax.swing.JMenu mTabSave;
    private javax.swing.JMenuItem miConfigureErrorModel;
    private javax.swing.JCheckBoxMenuItem miDrawAnchors;
    private javax.swing.JMenuItem miEditAnchors;
    private javax.swing.JMenuItem miTabBlank;
    private javax.swing.JMenuItem miTabClone;
    private javax.swing.JMenuItem miTabCloseAll;
    private javax.swing.JMenuItem miTabSave;
    private javax.swing.JMenuItem miTabSaveAll;
    private javax.swing.JMenuItem miTestErrorModel;
    private javax.swing.JSpinner spHeight;
    private javax.swing.JSpinner spRuns;
    private javax.swing.JSpinner spWidth;
    private javax.swing.JPopupMenu tabPopupMenu;
    private javax.swing.JTabbedPane tabbedPane;
    private javax.swing.JPanel toolBarContainer;
    private javax.swing.JPanel toolbar;
    private javax.swing.JPanel toolbar2;
    // End of variables declaration//GEN-END:variables

}
